"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.query_relations = exports.update_relations = exports.openfga_deinit = exports.openfga_init = exports.fix_object_without_colon = exports.openfgaOpaData = void 0;
const service_common_1 = require("@crosslab/service-common");
const sdk_1 = require("@openfga/sdk");
const syntax_transformer_1 = require("@openfga/syntax-transformer");
const child_process_1 = require("child_process");
const fs_1 = require("fs");
const readline_1 = __importDefault(require("readline"));
const datasource_1 = require("./database/datasource");
const model_1 = require("./database/model");
const encoding_1 = require("./encoding");
let fgaClient;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
let openfgaProcess;
const relationRepository = datasource_1.ApplicationDataSource.getRepository(model_1.RelationModel);
function fix_object_without_colon(object) {
    if (!object.includes(':'))
        object = object + ':~~~TYPE~~~';
    return object;
}
exports.fix_object_without_colon = fix_object_without_colon;
async function openfga_init() {
    fgaClient = new sdk_1.OpenFgaClient({
        apiScheme: 'http',
        apiHost: 'localhost:3013',
    });
    let readyPromiseResolve;
    const readyPromise = new Promise((resolve, _reject) => {
        readyPromiseResolve = resolve;
    });
    openfgaProcess = (0, child_process_1.spawn)('openfga', [
        'run',
        '--log-format',
        'json',
        '--playground-enabled=false',
        '--grpc-addr',
        'localhost:3012',
        '--http-addr',
        'localhost:3013',
    ], { stdio: ['ignore', 'pipe', 'pipe'] });
    readline_1.default
        .createInterface({
        input: openfgaProcess.stderr,
        terminal: false,
    })
        .on('line', line => {
        try {
            const json = JSON.parse(line.toString());
            if (json.msg.startsWith('HTTP server listening'))
                readyPromiseResolve();
            service_common_1.logger.log(json.level, json.msg, {
                ...json,
                msg: undefined,
                subsystem: 'openfga',
            });
        }
        catch (e) {
            service_common_1.logger.error('Failed to parse openfga log line', {
                line: line.toString(),
                error: e,
            });
        }
    });
    process.on('SIGINT', () => {
        openfgaProcess.kill('SIGINT');
        process.exit(0);
    });
    process.on('SIGTERM', () => {
        openfgaProcess.kill('SIGTERM');
        process.exit(0);
    });
    process.on('SIGQUIT', () => {
        openfgaProcess.kill('SIGQUIT');
        process.exit(0);
    });
    await readyPromise;
    const storeId = (await fgaClient.createStore({ name: 'crosslab' })).id;
    if (storeId === undefined)
        throw new Error('Store ID is undefined');
    fgaClient.storeId = storeId;
    const relation_model = (0, fs_1.readFileSync)('relation_model.fga', 'utf8');
    const authorization_model_id = (await fgaClient.writeAuthorizationModel((0, syntax_transformer_1.friendlySyntaxToApiSyntax)(relation_model))).authorization_model_id;
    if (authorization_model_id === undefined)
        throw new Error('Authorization model ID is undefined');
    fgaClient.authorizationModelId = authorization_model_id;
    exports.openfgaOpaData = {
        openfga: {
            store: storeId,
            authorization_model_id: authorization_model_id,
        },
    };
    await rehydrate();
}
exports.openfga_init = openfga_init;
function openfga_deinit() {
    openfgaProcess.kill(9);
}
exports.openfga_deinit = openfga_deinit;
async function rehydrate() {
    const relations = await relationRepository.find();
    if (relations.length === 0)
        return;
    await fgaClient.write({
        writes: relations.map((tuple) => ({
            user: (0, encoding_1.encode_object)(tuple.subject),
            relation: tuple.relation,
            object: fix_object_without_colon((0, encoding_1.encode_object)(tuple.object)),
        })),
    }, { transaction: { disable: true } });
}
async function update_relations(add, remove) {
    await relationRepository.createQueryBuilder().insert().values(add).orIgnore().execute();
    await relationRepository.remove(remove.map((tuple) => ({
        subject: tuple.subject,
        relation: tuple.relation,
        object: tuple.object,
    })));
    await fgaClient.write({
        writes: add.map((tuple) => ({
            user: (0, encoding_1.encode_object)(tuple.subject),
            relation: tuple.relation,
            object: fix_object_without_colon((0, encoding_1.encode_object)(tuple.object)),
        })),
        deletes: remove.map((tuple) => ({
            user: (0, encoding_1.encode_object)(tuple.subject),
            relation: tuple.relation,
            object: fix_object_without_colon((0, encoding_1.encode_object)(tuple.object)),
        })),
    }, { transaction: { disable: true } });
}
exports.update_relations = update_relations;
async function query_relations(subject, relation, object) {
    if (!object.includes(':'))
        object = object + ':';
    let tuples = [];
    let continuationToken = undefined;
    do {
        const result = await fgaClient.read({
            user: (0, encoding_1.encode_object)(subject),
            relation: relation,
            object: (0, encoding_1.encode_object)(object),
        }, { continuationToken });
        const local_tuples = result.tuples?.map(tuple => ({
            subject: (0, encoding_1.decode_object)(tuple.key.user),
            relation: tuple.key.relation,
            object: (0, encoding_1.decode_object)(tuple.key.object),
        })) ?? [];
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        tuples = tuples.concat(local_tuples);
        continuationToken =
            result.continuation_token === '' ? undefined : result.continuation_token;
    } while (continuationToken !== undefined);
    return tuples;
}
exports.query_relations = query_relations;
//# sourceMappingURL=openfga.js.map